
import { _decorator, Component, RenderTexture, Node, SpriteFrame, Camera, UITransform, gfx, Vec2, __private, ImageAsset, Texture2D, Layers } from 'cc';


export default class RenderUtil {

    /**
     * 获取像素数据
     * @param node 节点
     * @param flipY 垂直翻转数据
     */
    public static getPixelsData(node: Node, flipY: boolean = true) {
        if (!node.isValid) {
            return null;
        }
        // 节点宽高
        const width = Math.floor(node.getComponent(UITransform).width),
            height = Math.floor(node.getComponent(UITransform).height);
        // 创建临时摄像机用于渲染目标节点
        const cameraNode = new Node("cameraNode");
        cameraNode.layer = Layers.Enum.UI_2D;  //渲染层
        cameraNode.parent = node;
        const camera = cameraNode.addComponent(Camera);
        // camera.clearFlags |= Camera.ClearFlag.SOLID_COLOR;
        // camera.clearColor = color(0, 0, 0, 0);
        // 将节点渲染到 RenderTexture 中
        const renderTexture = new RenderTexture();
        renderTexture.initialize({
            width: width,
            height: height,
            // colorFormat: RenderTexture.PixelFormat.RGBA8888,
            // depthStencilFormat: RenderTexture.DepthStencilFormat.DEPTH_24_STENCIL_8
        });
        camera.targetTexture = renderTexture;
        // camera.render(node);
        // 获取像素数据
        const pixelsData = renderTexture.readPixels();
        // 销毁临时对象并返回数据
        renderTexture.destroy();
        cameraNode.destroy();
        // 垂直翻转数据
        if (flipY) {
            const length = pixelsData.length,
                lineWidth = width * 4,
                data = new Uint8Array(length);
            for (let i = 0, j = length - lineWidth; i < length; i += lineWidth, j -= lineWidth) {
                for (let k = 0; k < lineWidth; k++) {
                    data[i + k] = pixelsData[j + k];
                }
            }
            return data;
        }
        return pixelsData;
    }



    /**
     * 读取渲染纹理像素信息
     * @param texture 
     * @param filpY 是否翻转Y轴，默认true
     */
    public static readPixels(texture: __private._cocos_asset_assets_texture_base__TextureBase, filpY?: boolean): Uint8Array {
        // 通用版本
        var { width, height } = texture, gfxTexture = texture.getGFXTexture(), gfxDevice = texture['_getGFXDevice'](), bufferViews = [],
            region = new gfx.BufferTextureCopy, buffer = new Uint8Array(width * height * 4);
        // region.texOffset.x = region.texOffset.y = 0;
        region.texExtent.width = width;
        region.texExtent.height = height;
        bufferViews.push(buffer);
        gfxDevice?.copyTextureToBuffers(gfxTexture, bufferViews, [region]);
        // 翻转
        if (filpY !== false) {
            let i = 0, len1 = height / 2, len2 = width * 4, j: number, idx0: number, idx1: number;
            while (i < len1) {
                j = 0;
                while (j < len2) {
                    idx0 = i * len2 + j;
                    idx1 = (height - i - 1) * len2 + j++;
                    [buffer[idx0], buffer[idx1]] = [buffer[idx1], buffer[idx0]];
                }
                i++;
            }
        }
        return buffer;
        // 3.4及以上版本，有点问题
        // return texture.readPixels();
    }


    /**点击判断 射线
     * @param p=点击坐标
     * @param poly=检测范围顶点坐标
    */
    public static rayCasting(p: Vec2, poly: Array<Vec2>) {
        var px = p.x,
            py = p.y,
            flag = false

        for (var i = 0, l = poly.length, j = l - 1; i < l; j = i, i++) {
            var sx = poly[i].x,
                sy = poly[i].y,
                tx = poly[j].x,
                ty = poly[j].y

            // 点与多边形顶点重合  
            if ((sx === px && sy === py) || (tx === px && ty === py)) {
                return 'on'
            }

            // 判断线段两端点是否在射线两侧  
            if ((sy < py && ty >= py) || (sy >= py && ty < py)) {
                // 线段上与射线 Y 坐标相同的点的 X 坐标  
                var x = sx + (py - sy) * (tx - sx) / (ty - sy)

                // 点在多边形的边上  
                if (x === px) {
                    return 'on'
                }

                // 射线穿过多边形的边界  
                if (x > px) {
                    flag = !flag
                }
            }
        }

        // 射线穿过多边形边界的次数为奇数时点在多边形内  
        let result = flag ? 'in' : 'out';
        console.log('result = ' + result);
        return result;
    }


}
